require("RouteManager.nut");
require("Route.nut");

require("TownManager.nut");
require("BuildManager.nut");
require("Utils.nut");

require("WrightAI.nut");

class LuDiAIAfterFix extends AIController {
    MAX_TOWN_VEHICLES = AIGameSettings.GetValue("max_roadveh");
    MIN_DISTANCE = 40;
    MAX_DISTANCE = 115;
    MAX_DISTANCE_INCREASE = 138;

    townManager = null;
    routeManager = null;
    buildManager = null;
	scheduledRemovals = AIList();

    cargoClass = null;

    bestRoutesBuilt = null;
    allRoutesBuilt = null;

    wrightAI = null;
	
	loading = null;
	loadData = null;

    constructor() {
        townManager = TownManager();
        routeManager = RouteManager();
        buildManager = BuildManager();

        if(!AIController.GetSetting("select_town_cargo")) {
            cargoClass = AICargo.CC_PASSENGERS;
        }
        else {
            cargoClass = AICargo.CC_MAIL;
        }

        bestRoutesBuilt = false;
        allRoutesBuilt = false;

        wrightAI = WrightAI(cargoClass);
		
		loading = true;
    }

    function Start();
	
	function BuildRoadRoute(cityFrom) {
	    if ((routeManager.getRoadVehicleCount() < MAX_TOWN_VEHICLES - 10) && !allRoutesBuilt) {
			
	        local engineList = AIEngineList(AIVehicle.VT_ROAD);
			engineList.Valuate(AIEngine.IsValidEngine);
			engineList.KeepValue(1);
		    engineList.Valuate(AIEngine.IsBuildable);
		    engineList.KeepValue(1);
            engineList.Valuate(AIEngine.GetRoadType);
            engineList.KeepValue(AIRoad.ROADTYPE_ROAD);
            engineList.Valuate(AIEngine.CanRefitCargo, Utils.getCargoId(cargoClass));
            engineList.KeepValue(1);

			if (engineList.Count() != 0) {
			    engineList.Valuate(AIEngine.GetPrice);
				engineList.Sort(AIList.SORT_BY_VALUE, AIList.SORT_DESCENDING);
				
				local estimated_costs = 0;
				local engine_costs = (AIEngine.GetPrice(engineList.Begin()) + 500) * 2;
				local road_costs = AIRoad.GetBuildCost(AIRoad.ROADTYPE_ROAD, AIRoad.BT_ROAD) * 2 * MAX_DISTANCE_INCREASE;
				local clear_costs = AITile.GetBuildCost(AITile.BT_CLEAR_ROUGH) * MAX_DISTANCE_INCREASE;
				local station_costs = AIRoad.GetBuildCost(AIRoad.ROADTYPE_ROAD, cargoClass == AICargo.CC_PASSENGERS ? AIRoad.BT_BUS_STOP : AIRoad.BT_TRUCK_STOP) * 2;
				local depot_cost = AIRoad.GetBuildCost(AIRoad.ROADTYPE_ROAD, AIRoad.BT_DEPOT);
				estimated_costs += engine_costs + road_costs + clear_costs + station_costs + depot_cost;
//				AILog.Info("estimated_costs = " + estimated_costs + "; engine_costs = " + engine_costs + ", road_costs = " + road_costs + ", clear_costs = " + clear_costs + ", station_costs = " + station_costs + ", depot_cost = " + depot_cost);
                if (!Utils.HasMoney(estimated_costs)) {
				    return;
				}
			
				local articulatedList = AIList();
			    articulatedList.AddList(engineList);
			    articulatedList.Valuate(AIEngine.IsArticulated);
			    articulatedList.KeepValue(1);
			    local articulated = engineList.Count() == articulatedList.Count();
				
                if (cityFrom == null) {
                    cityFrom = townManager.getUnusedCity(bestRoutesBuilt, cargoClass);
                    if (cityFrom == null) {
                        if (AIController.GetSetting("pick_random")) {
                            townManager.m_usedCities.Clear();
                        } else {
                            if (!bestRoutesBuilt) {
                                bestRoutesBuilt = true;
                                townManager.m_usedCities.Clear();
//                              townManager.m_nearCityPairArray = [];
                                AILog.Warning("Best routes have been used! Year: " + AIDate.GetYear(AIDate.GetCurrentDate()));
                            } else {
                                allRoutesBuilt = true;
                                AILog.Warning("All routes have been used!");
                            }
                        }
                    }
                }

                if (cityFrom != null) {
	    		    AILog.Info("New city found: " + AITown.GetName(cityFrom));

                    townManager.findNearCities(cityFrom, MIN_DISTANCE, MAX_DISTANCE, bestRoutesBuilt, cargoClass);
                    if (!townManager.m_nearCityPairArray.len()) {
                        //second try with increased distance
                        townManager.findNearCities(cityFrom, MIN_DISTANCE, MAX_DISTANCE_INCREASE, bestRoutesBuilt, cargoClass);
                        if (!townManager.m_nearCityPairArray.len()) {
                            AILog.Info("No near city available");
                            cityFrom = null;
                        }
                    }
			    }

			    local cityTo = null;
                if (cityFrom != null) {
                    for (local i = 0; i < townManager.m_nearCityPairArray.len(); ++i) {
                        if (cityFrom == townManager.m_nearCityPairArray[i].m_cityFrom) {
                            if (!routeManager.townRouteExists(cityFrom, townManager.m_nearCityPairArray[i].m_cityTo)) {
                                cityTo = townManager.m_nearCityPairArray[i].m_cityTo;

                                if (routeManager.hasMaxStationCount(cityFrom, cityTo)) {
                                    cityTo = null;
                                    continue;
                                } else {
                                    break;
                                }
                            }
                        }
                    }

                    if (cityTo == null) {
                        cityFrom = null;
                    }
		    	}

                if (cityFrom != null && cityTo != null) {
			        AILog.Info("New near city found: " + AITown.GetName(cityTo));

			        local start_date = AIDate.GetCurrentDate();
                    local routeResult = routeManager.buildRoute(buildManager, cityFrom, cityTo, cargoClass, articulated);
                    if (routeResult) {
			            local days = AIDate.GetCurrentDate() - start_date;
                        AILog.Warning("Built route between: " + AITown.GetName(cityFrom) + " and " + AITown.GetName(cityTo) + " in " + days + " day" + (days != 1 ? "s" : "") + ".");
                    } else {
				        townManager.removeUsedCityPair(cityFrom, cityTo);
			            local days = AIDate.GetCurrentDate() - start_date;
				        AILog.Error(days + " day" + (days != 1 ? "s" : "") + " wasted!");
			        }

                    //cityFrom = cityTo; // use this line to look for a new town from the last town
                    cityFrom = null;
			    }
			}
        }
	}
	
	function RemoveLeftovers() {
	    if (scheduledRemovals.Count() == 0) {
		    return;
		}
		
		AIRoad.SetCurrentRoadType(AIRoad.ROADTYPE_ROAD);
		
		local clearedList = AIList();
		for (local tile = scheduledRemovals.Begin(); scheduledRemovals.HasNext(); tile = scheduledRemovals.Next()) {
		    if (AIRoad.IsRoadStationTile(tile) || AIRoad.IsDriveThroughRoadStationTile(tile)) {
			    if (TestRemoveRoadStation().TryRemove(tile)) {
				    clearedList.AddItem(tile, 0);
				}
			}
			else if (AIRoad.IsRoadDepotTile(tile)) {
			    if (TestRemoveRoadDepot().TryRemove(tile)) {
				    clearedList.AddItem(tile, 0);
				}
			}
			else {
			    // there was nothing to remove
				clearedList.AddItem(tile, 0);
			}
		}
		scheduledRemovals.RemoveList(clearedList);
	}

    function updateVehicles() {
	    local max_roadveh = AIGameSettings.GetValue("max_roadveh");
		if (max_roadveh != MAX_TOWN_VEHICLES) {
		    MAX_TOWN_VEHICLES = max_roadveh;
			AILog.Info("MAX_TOWN_VEHICLES = " + MAX_TOWN_VEHICLES);
		}

		for (local i = routeManager.m_townRouteArray.len() - 1; i >= 0; --i) {
            AIController.Sleep(1);

//			AILog.Info("managing route " + i + ". sellVehiclesInDepot");
            routeManager.m_townRouteArray[i].sellVehiclesInDepot();
			
//			AILog.Info("managing route " + i + ". sendNegativeProfitVehiclesToDepot");
            routeManager.m_townRouteArray[i].sendNegativeProfitVehiclesToDepot();


            if(MAX_TOWN_VEHICLES < routeManager.getRoadVehicleCount()) {
//			    AILog.Info("managing route " + i + ". sendLowProfitVehiclesToDepot");
                routeManager.m_townRouteArray[i].sendLowProfitVehiclesToDepot(routeManager);
            }

            if (MAX_TOWN_VEHICLES > routeManager.getRoadVehicleCount()) {
//              AILog.Info("managing route " + i + ". updateEngine");
                routeManager.m_townRouteArray[i].updateEngine();
				
//			    AILog.Info("managing route " + i + ". renewVehicles");
                routeManager.m_townRouteArray[i].renewVehicles();
				
//			    AILog.Info("managing route " + i + ". addVehicleToRoute");
                routeManager.m_townRouteArray[i].addVehicleToRoute();
			}

			local cityFrom = routeManager.m_townRouteArray[i].m_cityFrom;
			local cityTo = routeManager.m_townRouteArray[i].m_cityTo;            
			if (MAX_TOWN_VEHICLES > routeManager.getRoadVehicleCount()) {
			    local cityFrom = routeManager.m_townRouteArray[i].m_cityFrom;
			    local cityTo = routeManager.m_townRouteArray[i].m_cityTo;
//			    AILog.Info("managing route " + i + ". expandStations");
                if(AIController.GetSetting("station_spread") && AIGameSettings.GetValue("distant_join_stations") &&
			        routeManager.m_townRouteArray[i].expandStations()) {
                    AILog.Info("Expanded stations in " + AITown.GetName(cityFrom) + " and " + AITown.GetName(cityTo));
                }
				
//			    AILog.Info("managing route " + i + ". updateBridges")
			    routeManager.m_townRouteArray[i].updateBridges();
			}
			
//			AILog.Info("managing route " + i + ". removeIfUnserviced");
			if (routeManager.m_townRouteArray[i].removeIfUnserviced()) {
			     routeManager.m_townRouteArray.remove(i);
				 townManager.removeUsedCityPair(cityFrom, cityTo);
		    }
        }
    }

    function Save() {
	    if (loading) {
		    AILog.Error("WARNING! AI didn't finish loading previously saved data. It will be saving partial data!")
		}
        AILog.Warning("Saving...");

        local table = {};
        table.rawset("town_manager", townManager.saveTownManager());
        table.rawset("route_manager", routeManager.saveRouteManager());
        table.rawset("build_manager", buildManager.saveBuildManager());
		
        local scheduledRemovalsTable = {};
		for (local tile = scheduledRemovals.Begin(), i = 0; scheduledRemovals.HasNext(); tile = scheduledRemovals.Next(), ++i) {
		    scheduledRemovalsTable.rawset(i, [tile, scheduledRemovals.GetValue(tile)]);
		}
		table.rawset("scheduled_removes", scheduledRemovalsTable);

        table.rawset("best_routes_built", bestRoutesBuilt);
        table.rawset("all_routes_built", allRoutesBuilt);

        table.rawset("wrightai", wrightAI.save());

        AILog.Warning("Saved!");

        return table;
    }

    function Load(version, data) {
	    loading = true;
		loadData = data;
        AILog.Warning("Loading...");
    }
}

function LuDiAIAfterFix::Start() {
	if (loading) {
        if (loadData != null) {
		    if(loadData.rawin("town_manager")) {
                townManager.loadTownManager(loadData.rawget("town_manager"));
            }

            if(loadData.rawin("route_manager")) {
                routeManager.loadRouteManager(loadData.rawget("route_manager"));
            }

            if(loadData.rawin("build_manager")) {
                buildManager.loadBuildManager(loadData.rawget("build_manager"));
            }

		    if(loadData.rawin("scheduled_removes")) {
		        local scheduledRemovalsTable = loadData.rawget("scheduled_removes");
		        local i = 0;
		        while(scheduledRemovalsTable.rawin(i)) {
		            local tile = scheduledRemovalsTable.rawget(i);
			        scheduledRemovals.AddItem(tile[0], tile[1]);
			        ++i;
			    }
		    }
		    AILog.Info("Loaded " + scheduledRemovals.Count() + " scheduled removals.");

            if(loadData.rawin("best_routes_built")) {
                bestRoutesBuilt = loadData.rawget("best_routes_built");
            }

            if(loadData.rawin("all_routes_built")) {
                allRoutesBuilt = loadData.rawget("all_routes_built");
            }

            if(loadData.rawin("wrightai")) {
                wrightAI.load(loadData.rawget("wrightai"));
            }

            AILog.Warning("Game loaded.");
			loadData = null;
		}
		loading = false;
    }

    if (buildManager.hasUnfinishedRoute()) {
	    if (!AIController.GetSetting("road_support")){
		    return;
		}
        local cityFrom = buildManager.m_cityFrom;
        local cityTo = buildManager.m_cityTo;
		AILog.Warning("Unfinished route between " + AITown.GetName(cityFrom) + " and " + AITown.GetName(cityTo) + " detected!");

        local routeResult = routeManager.buildRoute(buildManager, cityFrom, cityTo, buildManager.m_cargoClass, buildManager.m_articulated);

        if (routeResult) {
            AILog.Warning("Built unfinished route between: " + AITown.GetName(cityFrom) + " and " + AITown.GetName(cityTo));
        }
    }
	
    local cargoId = Utils.getCargoId(cargoClass);
	local cargostr = AICargo.GetCargoLabel(cargoId);
    if (!AICompany.SetName("LuDiAI AfterFix " + cargostr)) {
        local i = 2;
        while (!AICompany.SetName("LuDiAI AfterFix " + cargostr + " #" + i)) {
            ++i;
        }
    }

	local cityFrom = null;
    while (AIController.Sleep(1)) {
        //pay loan
        Utils.RepayLoan();
        RemoveLeftovers();
        updateVehicles();
		if (AIController.GetSetting("road_support")) {
		    BuildRoadRoute(cityFrom);
		}

        //****
        //**** WrightAI functions from Wright.nut (https://wiki.openttd.org/AI:WrightAI) with major modifications
        //****
        if (AIController.GetSetting("air_support")) {
            if (!AIController.GetSetting ("road_support") || AIGameSettings.IsDisabledVehicleType(AIVehicle.VT_ROAD) ||
			    (routeManager.getRoadVehicleCount() > 125 || MAX_TOWN_VEHICLES <= 125 && routeManager.getRoadVehicleCount() > WrightAI.GetAircraftCount() * 4) ||
                (AIDate.GetYear(AIDate.GetCurrentDate()) > 1955 || allRoutesBuilt) ||
				routeManager.getRoadVehicleCount() >= MAX_TOWN_VEHICLES - 10) {
                wrightAI.BuildAirRoute();
            }
        }
        //****
    }
}




